#!/usr/bin/env python3
##
## -----------------------------------------------------------------
##    This file is part of WAPT Software Deployment
##    Copyright (C) 2012 - 2020  Tranquil IT https://www.tranquil.it
##    All Rights Reserved.
##
##    WAPT helps systems administrators to efficiently deploy
##    setup, update and configure applications.
## ------------------------------------------------------------------
##
import os
import sys

from waptserver.config import __version__

import email.utils

from flask import request, send_from_directory, Blueprint

import hashlib
import json
import random
import re
import requests
import shutil
import stat
import subprocess
import time
import traceback
import uuid
import logging
import zlib

import waptserver.config

from waptserver.utils import *
from waptserver.model import *

from waptserver.common import make_response, make_response_from_exception
from waptserver.decorators import requires_auth, wapt_db_readonly, require_wapt_db

try:
    # i18n
    from flask_babel import Babel
    from flask_babel import gettext
except ImportError:
    gettext = (lambda s: s)
    Babel = None

_ = gettext

enterprise = Blueprint('enterprise', __name__)


def prepare_request_data(request):
    if request.headers.get("Content-Encoding") != "gzip":
        return
    data = zlib.decompress(request.data)
    request.data = data


@enterprise.route('/api/v3/normalization_import_softwares', methods=['GET'])
@requires_auth(['admin', 'run_report'])
@require_wapt_db
def normalization_import_softwares():
    sql = """
    INSERT INTO normalization ( original_name, key, banned, windows_update, created_on, updated_on )
    SELECT
	S.name              as original_name,
	S.key               as key,
	false               as banned,
	false               as windows_update,
        current_timestamp   as created_on,
        current_timestamp   as updated_on
    FROM
	normalization N
	RIGHT join ( SELECT name, key  FROM hostsoftwares GROUP BY name, key ) S ON (N.original_name = S.name) AND (N.key = S.KEY)
    WHERE
	N.original_name is NULL;
    """
    start_time = time.time()
    with wapt_db.atomic() as trans:
        try:
            c = wapt_db.execute_sql(sql)
            r = make_response(result=c.rowcount,  success=True, request_time=time.time() - start_time)
            return r
        except Exception as e:
            trans.rollback()
            return make_response_from_exception(e)


@enterprise.route('/api/v3/normalization_filter', methods=['POST'])
@requires_auth(['admin', 'run_report'])
@require_wapt_db
def normalization_filter():
    start_time = time.time()
    prepare_request_data(request)
    data = ujson.loads(request.data)
    try:
        filter_string = data.get("filter_string", "").lower()
        logger.debug("filter string = " + filter_string)
        data = Normalization.select(Normalization.original_name, Normalization.key, Normalization.normalized_name, Normalization.banned, Normalization.windows_update)
        if len(filter_string) > 0:
            data = data.where(
                fn.LOWER(Normalization.original_name).contains(filter_string) |
                fn.LOWER(Normalization.key).contains(filter_string) |
                fn.LOWER(Normalization.normalized_name).contains(filter_string)
            )

        data = data.order_by(Normalization.original_name)

        data = list(data.dicts())
        r = make_response(data, success=True, request_time=start_time - time.time())
        return r
    except Exception as e:
        return make_response_from_exception(e)


@enterprise.route('/api/v3/normalization_write', methods=['POST'])
@requires_auth(['admin', 'run_report'])
@require_wapt_db
def normalization_write():
    start_time = time.time()
    prepare_request_data(request)
    data = ujson.loads(request.data)
    with wapt_db.atomic() as trans:
        try:
            for l in data:
                n = Normalization()
                n.original_name = l['original_name']
                n.key = l['key']
                n.normalized_name = l['normalized_name']
                n.banned = l['banned']
                n.windows_update = l['windows_update']
                n.save()
            trans.commit()
            data = [len(data)]
            r = make_response(data, msg='normalization_write', success=True, request_time=start_time - time.time())
            return r
        except Exception as e:
            trans.rollback()
            return make_response_from_exception(e)


@enterprise.route('/api/v3/reporting_exec', methods=['GET'])
@requires_auth(['admin', 'run_report'])
@wapt_db_readonly
def reporting_exec():
    start_time = time.time()
    R = None
    with wapt_db.atomic() as trans:
        try:
            id = request.args.get("id", None)
            if id is None:
                raise Exception("Param id is recquired")

            query = ReportingQueries.select(ReportingQueries.query).where(ReportingQueries.id == id)
            query = list(query.dicts())
            if 0 == len(query):
                raise Exception("Query id %d not found" % (id))
            data = ReportingQueries.raw(query[0]["query"])
            data = list(data.dicts())
            R = make_response(data, success=True, request_time=time.time() - start_time)
        except Exception as e:
            R = make_response_from_exception(e)
    return R


@enterprise.route('/api/v3/reporting_exec_raw', methods=['GET', 'POST'])
@requires_auth(['admin', 'run_report'])
@wapt_db_readonly
def reporting_exec_raw():
    start_time = time.time()
    prepare_request_data(request)
    R = None
    with wapt_db.atomic() as trans:
        try:
            query = ujson.loads(request.data)
            if not query:
                raise Exception("Param query is recquired")

            sql = query['query']
            if not sql:
                raise Exception("Param query is recquired")

            data = ReportingQueries.raw(sql)
            data = list(data.dicts())
            R = make_response(data, success=True, request_time=time.time() - start_time)
        except Exception as e:
            trans.rollback()
            R = make_response_from_exception(e)
    return R


@enterprise.route('/api/v3/reporting_list', methods=['GET'])
@requires_auth(['admin', 'run_report'])
@wapt_db_readonly
def reporting_list():
    start_time = time.time()
    wapt_db_connect()
    try:
        data = ReportingQueries.select(ReportingQueries.id, ReportingQueries.name, ReportingQueries.query, ReportingQueries.settings).order_by(ReportingQueries.name)
        data = list(data.dicts())
        return make_response(data, success=True, request_time=time.time() - start_time)
    except Exception as e:
        return make_response_from_exception(e)


@enterprise.route('/api/v3/reporting_save', methods=['GET', 'POST'])
@requires_auth(['admin', 'edit_report'])
@require_wapt_db
def reporting_save():
    start_time = time.time()
    prepare_request_data(request)
    try:
        queries_list = ujson.loads(request.data)
        if not queries_list:
            raise Exception("reporting : No data supplied")
    except Exception as e:
        return make_response_from_exception(e)

    with wapt_db.atomic() as trans:
        try:
            data = []
            for query in queries_list:
                rq = ReportingQueries()
                if "id" in query:
                    if query["id"] >= 0:
                        rq.id = query["id"]
                if "name" in query:
                    rq.name = query["name"]
                if "query" in query:
                    rq.query = query["query"]
                if "settings" in query:
                    rq.settings = query["settings"]
                rq.save()
                data.append({query["id"]: rq.id})
            trans.commit()
            R = make_response(data, success=True, request_time=time.time() - start_time)
            return R
        except Exception as e:
            trans.rollback()
            return make_response_from_exception(e)


@enterprise.route('/api/v3/reporting_delete', methods=['GET', 'POST'])
@requires_auth(['admin', 'edit_report'])
@require_wapt_db
def reporting_delete():
    start_time = time.time()
    prepare_request_data(request)
    try:
        id_list = ujson.loads(request.data)
        if not id_list:
            raise Exception("reporting : No data supplied")
    except Exception as e:
        return make_response_from_exception(e)

    with wapt_db.atomic() as trans:
        try:
            data = []
            cnt = ReportingQueries.delete().where(ReportingQueries.id.in_(id_list)).execute()
            trans.commit()
            R = make_response(cnt, success=True, request_time=time.time() - start_time)
            return R
        except Exception as e:
            trans.rollback()
            return make_response_from_exception(e)


# Users

@enterprise.route('/api/v3/waptusers', methods=['GET', 'PUT'])
@requires_auth(['admin'])
@wapt_db_readonly
def waptusers_get():
    start_time = time.time()
    if request.method == 'GET':
        try:
            data = list(WaptUsers.select().dicts())
            R = make_response(data, success=True, request_time=time.time() - start_time, msg='%s Wapt users' % len(data))
            return R
        except Exception as e:
            return make_response_from_exception(e)
    elif request.method == 'PUT':
        # we have a single delta to apply
        prepare_request_data(request)
        try:
            delta = ujson.loads(request.data)
            if not delta:
                raise Exception("No delta rows list supplied")
        except Exception as e:
            return make_response_from_exception(e)

        with wapt_db.atomic() as trans:
            try:
                result = []
                for row in delta:
                    id = row.get('id')
                    if id is not None and id >= 0:
                        # get existing
                        user_rec = WaptUsers.get(id=id)
                    else:
                        # create new
                        user_rec = WaptUsers()
                    for (k, v) in row.items():
                        if k != 'id' and hasattr(WaptUsers, k):
                            setattr(user_rec, k, v)
                    user_rec.save()
                    result.append(user_rec)
                trans.commit()
                R = make_response(result, success=True, request_time=time.time() - start_time)
                return R
            except Exception as e:
                trans.rollback()
                return make_response_from_exception(e)


@enterprise.route('/api/v3/waptusers', methods=['POST'])
@requires_auth(['admin'])
@require_wapt_db
def waptusers_post():
    DataClass = WaptUsers
    start_time = time.time()
    prepare_request_data(request)
    try:
        changes = ujson.loads(request.data)
        if not changes:
            raise Exception("No changes data supplied")
    except Exception as e:
        return make_response_from_exception(e)

    with wapt_db.atomic() as trans:
        try:
            result = []
            for change in changes:
                update_type = change.get('update_type')
                id = change.get('id')
                old_values = change.get('old_values')
                new_values = change.get('new_values')

                # insert
                if update_type in (1, 2):
                    if id is not None and id >= 0:
                        # get existing
                        rec = DataClass.get(id=id)
                    else:
                        # create new
                        rec = DataClass(**new_values)
                    # init with new values
                    if new_values:
                        for (k, v) in new_values.items():
                            if k != 'id' and hasattr(DataClass, k):
                                setattr(rec, k, v)
                    new_id = rec.save()
                    result.append((id, new_id))
                # delete
                elif update_type == 3:
                    DataClass.delete_by_id(id)
                    result.append((id, None))

            trans.commit()
            R = make_response(result, success=True, request_time=time.time() - start_time)
            return R
        except Exception as e:
            trans.rollback()
            return make_response_from_exception(e)


@enterprise.route('/api/v3/waptusers_delete', methods=['POST'])
@requires_auth(['admin'])
@require_wapt_db
def waptusers_delete():
    start_time = time.time()
    prepare_request_data(request)
    try:
        id_list = ujson.loads(request.data)
        if not id_list:
            raise Exception("No data supplied")
    except Exception as e:
        return make_response_from_exception(e)

    with wapt_db.atomic() as trans:
        try:
            data = []
            cnt = WaptUsers.delete().where(WaptUsers.id.in_(id_list)).execute()
            trans.commit()
            return make_response(cnt, success=True, request_time=time.time() - start_time, msg='%s users deleted' % len(id_list))
        except Exception as e:
            trans.rollback()
            return make_response_from_exception(e)
