#!/usr/bin/env python3
##
## -----------------------------------------------------------------
##    This file is part of WAPT Software Deployment
##    Copyright (C) 2012 - 2020  Tranquil IT https://www.tranquil.it
##    All Rights Reserved.
##
##    WAPT helps systems administrators to efficiently deploy
##    setup, update and configure applications.
## ------------------------------------------------------------------
##

"""waptpython waptwua.py <action>

    Script which scans the computer for Windows Update based on wsusscn2 cab
    Stores the result of scan in waptdb
    Can download and apply Windows Updates from wapt server
    based on allowed_classifications, forbidden_updates, allowed_updates, allowed_severities list

    An update is allowed if :
      not in forbidden and
        (
          (in allowed_classifications if specified and in allowed_severities if specified)
          or
          in allowed_updates if specified
        )

    <action> can be :
        scan : updates wsusscn2.cab, and checks current computer againt allowed KB
        download : download necessary updates from wapt server and push them in cache
        install : install allowed cached updates
"""
__version__ = "1.9.0.0"


import csv
import datetime
import hashlib
import json
import logging
import os
import platform
import requests
import subprocess
import sys
import time
import glob

import pythoncom
from win32com.server.util import wrap
from win32com.server.policy import MappedWrapPolicy, DesignatedWrapPolicy, EventHandlerPolicy
from win32com.client import CastTo
import pywintypes

import random

from optparse import OptionParser
from urllib.parse import urlparse
from setuphelpers import *

from waptutils import jsondump, get_time_delta, int2uhex

from waptenterprise.waptwua.utils import *
from waptenterprise.waptwua.constants import *

from iniparse import RawConfigParser

logger = logging.getLogger('waptwua')

try:
    import wmi
except:
    logger.warning('WMI is not available, some WaptWUA features will not work')
    wmi = None


def jsondump_sorted(o, **kwargs):
    return jsondump(o, sort_keys=True, separators=(',', ':'), **kwargs)


def get_product_id(expr):
    """Find product ids matching expr"""
    result = []
    match = re.compile(expr, re.IGNORECASE)
    for key, value in Products.items():
        if match.match(value) or expr == key:
            result.append(key)
    return result


Severities = {
    None: 'None',
    0: 'Critical',
    1: 'Important',
    2: 'Moderate',
    3: 'Low',
}

InstallResult = {
    0: 'NotStarted',
    1: 'InProgress',
    2: 'Succeeded',
    3: 'SucceededWithErrors',
    4: 'Failed',
    5: 'Aborted',
}


WUA_MAJOR_VERSION = 7
WUA_MINOR_VERSION = 6

_startupinfo_hide = subprocess.STARTUPINFO()
_startupinfo_hide.wShowWindow = subprocess.SW_HIDE


def create_waptwua_tasks(wapt, default_categories=['CriticalUpdates', 'SecurityUpdates', 'DefinitionUpdates', 'UpdateRollups']):
    """Create periodic tasks for :
    - download of wsuscn2.scn and scan of pending update tasks
    - install of updates
    """
    waptpython_path = makepath(wapt.wapt_base_dir, 'waptpython.exe')
    waptwua_path = makepath(wapt.wapt_base_dir, 'waptwua')

    if task_exists('waptwua'):
        delete_task('waptwua')
    if task_exists('waptwua-scan'):
        delete_task('waptwua-scan')
    if task_exists('waptwua-install'):
        delete_task('waptwua-install')

     # randowmize a little the scan
    dt = datetime.datetime.now()+datetime.timedelta(hours=random.randrange(0, 5), minutes=random.randrange(0, 59))
    create_daily_task('waptwua-scan', waptpython_path, '"%s" download -C %s' % (makepath(waptwua_path, 'waptwua.py'), ','.join(default_categories)), start_hour=dt.hour, start_minute=dt.minute)
    create_daily_task('waptwua-install', waptpython_path, '"%s" install -C %s' % (makepath(waptwua_path, 'waptwua.py'), ','.join(default_categories)), start_hour=3, start_minute=0)


def remove_waptwua_tasks():
    if task_exists('waptwua-scan'):
        run_task
        delete_task('waptwua-scan')
    if task_exists('waptwua-install'):
        delete_task('waptwua-install')


class WAPTDiskSpaceException(Exception):
    def __init__(self, message, free_space):
        self.message = message
        self.free_space = free_space

    def __str__(self):
        return str(self.message) + '; free space: ' + str(self.free_space / 2**20) + 'MB'


class WaptJSonSettings(object):
    CRITERIA_LIST = []
    CRITERIA_BOOL = []
    CRITERIA_STR = []

    _class_key = 'waptsettings'

    def __init__(self, **kwargs):
        for (k, v) in kwargs.items():
            if k in self.CRITERIA_LIST+self.CRITERIA_BOOL+self.CRITERIA_STR:
                setattr(self, k, v)
            else:
                raise Exception('Unknown %s parameter %s' % (self.__class__.__name__, k))

    def load_from_options(self, options_parser):
        for key in self.CRITERIA_LIST:
            fqkey = '%s.%s' % (self._class_key, key)
            if getattr(options_parser, fqkey, None) is not None:
                setattr(self, key, ensure_list(getattr(options_parser, fqkey, None)))
        for key in self.CRITERIA_BOOL+self.CRITERIA_STR:
            fqkey = '%s.%s' % (self._class_key, key)
            if getattr(options_parser, fqkey, None) is not None:
                setattr(self, key, getattr(options_parser, fqkey, None))
        return self

    def load_from_ini(self, config_filename=None, config=None, section=None):
        if section is None:
            section = self._class_key
        if config is None:
            config = RawConfigParser()
            config.read(config_filename)
        if config.has_section(section):
            for key in self.CRITERIA_BOOL:
                if config.has_option(section, key):
                    setattr(self, key, config.getboolean(section, key))
            for key in self.CRITERIA_LIST:
                if config.has_option(section, key):
                    setattr(self, key, ensure_list(config.get(section, key)))
            for key in self.CRITERIA_STR:
                if config.has_option(section, key):
                    setattr(self, key, config.get(section, key))
        return self

    def merge_rules(self, rules_dict, ignore_unknown_keys=False):
        """Merge lists with same key name as mine from rules_dict
        Other item are just overwritten.
        Keys of rules_dict must

        Args:
            rules_dict (dict) : {key: (None or list or str or bool..)}

        Returns:
            None

        """

        for (k, v) in rules_dict.items():
            if k in self.CRITERIA_LIST+self.CRITERIA_BOOL+self.CRITERIA_STR:
                old_value = getattr(self, k)
                if old_value is None or not isinstance(old_value, list):
                    # overrides value even if already set ...
                    setattr(self, k, v)
                elif v is not None:
                    # append to lists
                    for item in v:
                        if not item in old_value:
                            old_value.append(item)
            elif not ignore_unknown_keys:
                raise Exception('Unknown parameter %s for %s' % (k, self.__class__.__name__))

    def merge_rules_from_files(self, json_filenames_glob=None):
        """Merge all json files matching glob.glob(json_filenames_glob)
        """
        for json_fn in glob.glob(json_filenames_glob):
            rules = json.load(open(json_fn, 'r'))
            self.merge_rules(rules)
        return self

    def save_to_ini(self, config_filename=None, config=None, section='waptwua'):
        updated = False
        if config is None:
            config = RawConfigParser()
            config.read(config_filename)
            temp_ini = True

        def _encode_ini_value(value, key=None):
            if isinstance(value, list):
                return ','.join(value)
            elif value is None:
                return ''
            else:
                return value

        for key in self.CRITERIA_LIST+self.CRITERIA_BOOL+self.CRITERIA_STR:
            value = getattr(self, key, None)
            ini_key = key
            if value is None:
                if config.has_option(section, ini_key):
                    config.remove_option(section, ini_key)
                    updated = True
            else:
                if config.has_section(section) and config.has_option(section, ini_key):
                    if key in self.CRITERIA_LIST:
                        old_value = ensure_list(config.get(section, ini_key))
                    elif key in self.CRITERIA_BOOL:
                        old_value = config.getboolean(section, ini_key)
                    else:
                        old_value = config.get(section, ini_key)
                else:
                    old_value = None

                if old_value != value:
                    if not config.has_section(section):
                        config.add_section(section)
                    if key in self.CRITERIA_LIST:
                        config.set(section, ini_key, ','.join(value))
                    else:
                        config.set(section, ini_key, value)
                    updated = True

        if temp_ini and updated:
            config.write(open(config_filename, 'wb'))
        return updated

    def load_from_waptdb(self, wapt):
        rules = wapt.read_param(self._class_key, ptype='json')
        if rules:
            for key in self.CRITERIA_LIST+self.CRITERIA_BOOL+self.CRITERIA_STR:
                if getattr(self, key, None) is None:
                    setattr(self, key, rules.get(key))
        return self

    def save_to_waptdb(self, wapt):
        wapt.write_param(self._class_key, self.as_dict())

    def as_dict(self):
        return {key: getattr(self, key) for key in self.CRITERIA_LIST+self.CRITERIA_BOOL+self.CRITERIA_STR}

    def __repr__(self):
        return repr(self.as_dict())


class WaptWUARules(WaptJSonSettings):
    CRITERIA_LIST = ['allowed_products', 'allowed_classifications', 'allowed_severities',
                     'allowed_updates', 'forbidden_updates', 'allowed_kbs', 'forbidden_kbs']
    CRITERIA_BOOL = ['default_allow']
    CRITERIA_STR = []

    _class_key = 'waptwua_rules'

    def __init__(self, **kwargs):
        self.allowed_products = None
        self.allowed_classifications = None
        self.allowed_severities = None
        self.allowed_updates = None
        self.forbidden_updates = None
        self.allowed_kbs = None
        self.forbidden_kbs = None
        self.default_allow = None

        super(WaptWUARules, self).__init__(**kwargs)

    def is_allowed(self, update):
        """Check if an update is allowed
            allowed if not explicitly forbidden and in allowed classifications, products and criticities or list of update ids

        Args:
            update (IUpdate)

        Returns:
            bool
        """
        update_id = "%s_%s" % (update.Identity.UpdateID, update.Identity.RevisionNumber)

        if self.forbidden_updates is not None and update_id in self.forbidden_updates:
            return False

        # check by KB list as well as by updateId list
        kbs = ["KB%s" % kb for kb in update.KBArticleIDs]

        if self.forbidden_kbs is not None:
            for kb in kbs:
                if kb in self.forbidden_kbs:
                    return False

        # individual update_id_revision allowance
        if self.allowed_updates is not None and update_id in self.allowed_updates:
            return True

        # individual KB allowance
        if self.allowed_kbs is not None:
            for kb in kbs:
                if kb in self.allowed_kbs:
                    return True

        # product filter
        if self.allowed_products is not None:
            products = [c.Name for c in update.Categories if c.Type == 'Product']
            if products:
                product = products[0]
            else:
                product = ''

            if product in self.allowed_products:
                return True

        # severities filter
        if self.allowed_severities is not None:
            if update.MsrcSeverity in self.allowed_severities:
                return True

        # classification filter
        if self.allowed_classifications is not None:
            # get updateClassification list of this update
            update_class = [c.Name for c in update.Categories if c.Type == 'UpdateClassification']
            for cat in update_class:
                if cat in self.allowed_classifications:
                    return True

        return self.default_allow


class WaptWUAParams(WaptJSonSettings):
    CRITERIA_LIST = []
    CRITERIA_BOOL = ['direct_download', 'default_allow']
    CRITERIA_STR = ['filter', 'download_scheduling', 'install_scheduling', 'install_delay', 'postboot_delay']

    _class_key = 'waptwua_params'

    def __init__(self, **kwargs):
        self._filter = None

        self.direct_download = False
        self.default_allow = False
        self.download_scheduling = None
        self.install_scheduling = None
        self.install_delay = None
        self.postboot_delay = '10m' # delay the download / install task after boot

        super(WaptWUAParams, self).__init__(**kwargs)

    @property
    def filter(self):
        if self._filter is not None:
            return self._filter
        else:
            return "Type='Software' or Type='Driver'"

    @filter.setter
    def filter(self, value):
        self._filter = value

    def is_delayed(self, update):
        # don't install before install_delay after publishing date
        if self.install_delay is not None:
            try:
                delay_delta = get_time_delta(self.install_delay, 'd')
            except Exception as e:
                logger.critical('Bad time install_delay: %s %s' % (self.install_delay, repr(e)))
                return True
            changetime = datetime.datetime.fromtimestamp(timestamp=update.LastDeploymentChangeTime.timestamp(), tz=update.LastDeploymentChangeTime.tzinfo)
            delayed = datetime.datetime.now(changetime.tzinfo) < (changetime + delay_delta)
        else:
            delayed = False
        return delayed


class WaptWUA(EnsureWUAUServRunning):
    def __init__(self, wapt, params=None, windows_updates_rules=None):
        """Initialize a waptwua client

        Args:
            windows_updates_rules (WaptWUARules):  {'allowed_products','forbidden_updates','allowed_updates','allowed_severities','allowed_classifications'}

        """
        self.wapt = wapt
        self.cache_path = os.path.abspath(makepath(wapt.wapt_base_dir, 'waptwua', 'cache'))
        self.wsusscn2 = makepath(self.cache_path, 'wsusscn2.cab')
        self._update_session = None
        self._update_service_manager = None
        self._update_searcher = None
        self._offline_update_searcher = None
        self._update_service = None
        self.wapttask = None
        self.rules_packages = None

        #
        self._updates = None
        # to store successful changes in read only properties of _updates after initial scan
        self._cached_updates = {}

        self._installed_updates = None

        if not params:
            params = WaptWUAParams()
            if self.wapt:
                if self.wapt.config and self.wapt.config.has_section('waptwua'):
                    params.load_from_ini(config=self.wapt.config, section='waptwua')
                # else:
                #    windows_updates_rules.load_from_waptdb(self.wapt)
        self.params = params

        if windows_updates_rules:
            self.windows_updates_rules = windows_updates_rules
        else:
            self.load_windows_updates_rules_from_packages()

        self.windows_updates_rules.default_allow = self.params.default_allow

    def load_windows_updates_rules_from_packages(self):
        """Load list of wsus installed packages and load asociated json windows_rules files

        """
        self.windows_updates_rules = WaptWUARules()
        self.windows_updates_rules.load_from_ini(config=self.wapt.config, section='waptwua')
        self.rules_packages = self.wapt.waptdb.query("""select package_uuid,package,version,persistent_dir from wapt_localstatus where section='wsus' and install_status='OK' order by install_date""")
        for p in self.rules_packages:
            if p.get('persistent_dir') and os.path.isdir(p.get('persistent_dir')):
                self.windows_updates_rules.merge_rules_from_files(makepath(p['persistent_dir'], 'waptwua_rules.json'))
            else:
                logger.critical('Bad persistent dir for package %s' % p.get('package'))

    def apply_waptwua_settings_to_host(self):
        """Apply waptwua service specific settings
        """
        try:
            # check waptwua
            if self.wapt.waptwua_enabled:
                print('Disabling Windows auto update service, using WaptWUA instead')
                self.disable_ms_windows_update_service()
            elif self.wapt.waptwua_enabled is not None and not self.wapt.waptwua_enabled:
                print('Enabling Windows update service')
                self.enable_ms_windows_update_service()
        except Exception as e:
            logger.critical('Unable to set waptwua policies : %s' % e)

    def wuarepo(self):
        return self.wapt.wua_repository

    def ensure_minimum_wua_version(self):
        if wua_agent_version() < Version('7.6.7601.23435'):
            raise EWUABadSystem("first use fixwua before standard updating method")
        return True

    def is_allowed(self, update):
        return self.windows_updates_rules.is_allowed(update)

    def cached_update_property(self, update, key, safe=False):
        try:
            if hasattr(update, key):
                update_id = "%s_%s" % (update.Identity.UpdateID, update.Identity.RevisionNumber)
                if update_id in self._cached_updates and key in self._cached_updates[update_id]:
                    return self._cached_updates[update_id][key]
                else:
                    return getattr(update, key, None)
            else:
                return getattr(update, key, None)
        except Exception as e:
            if safe:
                return '???'
            else:
                raise e

    def store_cached_update_property(self, update, key, value):
        update_id = "%s_%s" % (update.Identity.UpdateID, update.Identity.RevisionNumber)
        if not update_id in self._cached_updates:
            self._cached_updates[update_id] = {}
        cached = self._cached_updates[update_id]
        cached[key] = value

    def update_as_dict(self, update):
        """Convert a IUpdate instance into a dict

        Args:
            update (IUpdate)

        Returns:
            dict
        """
        if hasattr(update, 'Categories'):
            products = [c.Name for c in update.Categories if c.Type == 'Product']
            classifications = [c.Name for c in update.Categories if c.Type == 'UpdateClassification']
            # https://docs.microsoft.com/en-us/windows/desktop/api/wuapi/nn-wuapi-iupdate
            result = dict(
                update_id="%s_%s" % (update.Identity.UpdateID, update.Identity.RevisionNumber),
                languages=["%s" % l for l in update.Languages],
                title=update.Title,
                update_type=UpdateType.get(update.Type, update.Type),
                kbids=["%s" % kb for kb in update.KBArticleIDs],
                severity=update.MsrcSeverity,
                changetime=datetime2isodate(datetime.datetime.fromtimestamp(timestamp=update.LastDeploymentChangeTime.timestamp(), tz=update.LastDeploymentChangeTime.tzinfo)),
                product=(products and products[0]) or "",
                classification=(classifications and classifications[0] or ""),
                download_urls=self.get_downloads_for_update(update),
                min_download_size=int(update.MinDownloadSize),
                max_download_size=int(update.MaxDownloadSize),
                superseded_update_ids=["%s" % id for id in update.SupersededUpdateIDs],
                security_bulletin_ids=["%s" % id for id in update.SecurityBulletinIDs],
                is_mandatory=update.IsMandatory,
                reboot_behaviour=InstallationRebootBehavior.get(update.InstallationBehavior.RebootBehavior, update.InstallationBehavior.RebootBehavior),
                installation_impact=InstallationImpact.get(update.InstallationBehavior and update.InstallationBehavior.Impact, update.InstallationBehavior and update.InstallationBehavior.Impact),
                uninstallation_impact=InstallationImpact.get(update.UninstallationBehavior and update.UninstallationBehavior.Impact, update.UninstallationBehavior and update.UninstallationBehavior.Impact),
                can_request_user_input=update.InstallationBehavior.CanRequestUserInput,
                requires_network_connectivity=update.InstallationBehavior and update.InstallationBehavior.RequiresNetworkConnectivity,
                is_beta=update.IsBeta,
                is_uninstallable=update.IsUninstallable,
                uninstallation_notes=update.UninstallationNotes,
                support_url=update.SupportUrl,
                release_notes=update.ReleaseNotes,
            )

        else:
            result = dict(
                update_id="%s_%s" % (update.Identity.UpdateID, update.Identity.RevisionNumber),
                languages=None,
                title=update.Title,
                update_type="",
                kbids=[],
                severity="",
                changetime=datetime2isodate(datetime.datetime.fromtimestamp(int(bytes(update.Date)))),
                product="",
                classification="",
                download_urls=None,
                min_download_size=None,
                max_download_size=None,
                superseded_update_ids=None,
                security_bulletin_ids=None,
                is_mandatory=None,
                reboot_behaviour=None,
                can_request_user_input=None,
                requires_network_connectivity=None,
                is_beta=None,
                is_uninstallable=None,
                installation_impact=None,
                uninstallation_impact=None,
                uninstallation_notes=None,
                support_url=None,
                release_notes=None,
            )

        return result

    def update_local_status_as_dict(self, update, updates_history=None):
        """Convert a IUpdate instance into a dict

        Args:
            update (IUpdate)

        Returns:
            dict
        """
        result = dict(
            update_id="%s_%s" % (update.Identity.UpdateID, update.Identity.RevisionNumber),
            installed=self.cached_update_property(update, 'IsInstalled', True),
            present=self.cached_update_property(update, 'IsPresent', True),
            hidden=self.cached_update_property(update, 'IsHidden', True),
            downloaded=self.cached_update_property(update, 'IsDownloaded', True),
            allowed=self.windows_updates_rules.is_allowed(update),
            history=[],
        )

        if updates_history:
            for e in sorted(updates_history, key=lambda h: h['date']):
                if e['update_id'] == result['update_id']:
                    result['history'].append(e)

        status = None
        install_date = None

        if result['history']:
            e = result['history'][-1]
            if e.get('operation') == 1:
                if install_date is None or (e.get('date') and e.get('date') > install_date):
                    install_date = e.get('date')  # install
                if e['result_code'] == 'Failed' and status is None:
                    status = 'ERROR'
                if e['result_code'] == 'Succeeded' and (status is None or status == 'ERROR'):
                    status = 'OK'

        delayed = self.params.is_delayed(update)

        if status is None:
            if self.cached_update_property(update, 'IsInstalled', True):
                status = 'OK'
            elif not self.cached_update_property(update, 'IsInstalled', True) and not update.IsHidden:
                status = 'PENDING'
            else:
                status = 'DISCARDED'

        result['delayed'] = delayed
        result['status'] = status

        return result

    def update_history(self):
        """Returns Windows updates installs history

        Args:

        Returns:
            list of dict

        """
        searcher = self.update_searcher()
        history_count = searcher.GetTotalHistoryCount()

        result = None

        if history_count > 0:
            result = []
            for entry in searcher.QueryHistory(0, history_count):
                entry2 = CastTo(entry, 'IUpdateHistoryEntry2')
                result.append(dict(
                    operation=entry2.Operation,
                    update_id="%s_%s" % (entry2.UpdateIdentity.UpdateID, entry2.UpdateIdentity.RevisionNumber),
                    date=datetime2isodate(datetime.datetime.fromtimestamp(timestamp=entry2.Date.timestamp(), tz=entry2.Date.tzinfo)),
                    title=entry2.Title,
                    result_code=OperationResultCode.get(entry2.ResultCode, entry2.ResultCode),
                    application_id=entry2.ClientApplicationID,
                    description=entry2.Description,
                    categories=[c.Name for c in entry2.Categories],
                ))

            result.sort(key=lambda a: a['update_id'])
        return result

    def update_session(self):
        """https://msdn.microsoft.com/fr-fr/library/windows/desktop/aa386854(v=vs.85).aspx

        Returns:
            IUpdateSession
        """
        if self._update_session is None:
            self._update_session = CastTo(win32com_ensure_dispatch_patch("Microsoft.Update.Session"), 'IUpdateSession3')

        return self._update_session

    def update_service_manager(self):
        """https://msdn.microsoft.com/fr-fr/library/windows/desktop/aa386819(v=vs.85).aspx

        Returns:
            IUpdateServiceManager
        """
        if self._update_service_manager is None:
            self._update_service_manager = CastTo(self.update_session().CreateUpdateServiceManager(), 'IUpdateServiceManager2')
        return self._update_service_manager

    def download_wsusscan_cab(self):
        """Download from wapt server the last version of wsusscn2.cab database for offline update scan.
        """
        repo = self.wuarepo()
        if repo:
            with repo.get_requests_session() as session:
                try:
                    if self.params.direct_download or ('download.windowsupdate.com' in repo.repo_url):
                        cab_location = 'http://download.windowsupdate.com/microsoftupdate/v6/wsusscan/wsusscn2.cab'
                    else:
                        cab_location = '%swua/wsusscn2.cab' % repo.repo_url
                    cab_target = self.wsusscn2
                    cab_current_date = ensure_unicode(self.wapt.read_param('waptwua.wsusscn2cab_date'))
                    r = session.head(
                        cab_location,
                        timeout=self.wuarepo().timeout,
                        allow_redirects=True,
                    )
                    r.raise_for_status()
                    cab_new_date = httpdatetime2isodate(r.headers['last-modified'])
                    if not cab_current_date or not isfile(cab_target) or (cab_new_date > cab_current_date):
                        print('Downloading wsusscn2.cab file from %s' % cab_location)
                        if self.params.direct_download or ('download.windowsupdate.com' in repo.repo_url):
                            wget('http://download.windowsupdate.com/microsoftupdate/v6/wsusscan/wsusscn2.cab', cab_target, connect_timeout=self.wuarepo().timeout, requests_session=session, limit_bandwidth=self.wapt.limit_bandwidth)
                        else:
                            sumwsusscn2 = wgets('%s.sha1' % cab_location, requests_session=session)
                            wget(cab_location, cab_target, connect_timeout=self.wuarepo().timeout, requests_session=session, sha1=sumwsusscn2.decode('utf8'), limit_bandwidth=self.wapt.limit_bandwidth)
                        if isfile(makepath(self.wapt.wapt_base_dir, 'utils', 'signtool.exe')):
                            subprocess.check_output(r'"%s" verify /pa "%s"' % (makepath(self.wapt.wapt_base_dir, 'utils', 'signtool.exe'), cab_target))
                        self.wapt.write_param('waptwua.wsusscn2cab_date', cab_new_date)
                        logger.debug('New wusscn2.cab date : %s' % cab_new_date)
                    return cab_new_date
                except Exception as e:
                    self.wapt.write_param('waptwua.status', 'ERROR DOWNLOADING WSUSSCAN')
                    raise e

    def update_wsusscan_cab(self):
        try:
            return self.download_wsusscan_cab()
        except Exception as e:
            if isfile(self.wsusscn2):
                logger.error('Unable to refresh wsusscan cab, using old one. (error: %s)' % ensure_unicode(e))
            else:
                logger.error('Unable to get wsusscan cab, aborting.')
                raise

    def update_searcher(self):
        """Instantiate a updateSearcher instance
        https://msdn.microsoft.com/en-us/library/windows/desktop/aa386515(v=vs.85).aspx

        Returns:
            IUpdateSearcher

        """
        # use wsus offline updates index cab
        if not self._offline_update_searcher:
            print('   Connecting to local update searcher using offline wsusscn2 file...')
            if not os.path.isfile(self.wsusscn2):
                self.download_wsusscan_cab()
            self._update_service = self.update_service_manager().AddScanPackageService("Offline Sync Service", self.wsusscn2)
            self._offline_update_searcher = CastTo(self.update_session().CreateUpdateSearcher(), 'IUpdateSearcher3')
            # use offline only
            self._offline_update_searcher.ServerSelection = 3  # other
            self._offline_update_searcher.ServiceID = self._update_service.ServiceID
            print('   Offline Update searcher ready...')
        return self._offline_update_searcher

    def updates(self):
        """List of current updates scanned againts wsusscn2 cab and computer
        https://msdn.microsoft.com/en-us/library/windows/desktop/aa386099(v=vs.85).aspx

        Returns:
            list of IUpdate
        """
        if self._updates is None:
            self.wapt.write_param('waptwua.status', 'SCANNING')
            try:
                print('Looking for updates with filter: %s' % self.params.filter)
                IID_ISearchCompletedCallback = '{88AEE058-D4B0-4725-A2F1-814A67AE964C}'

                class SearchCallback(MappedWrapPolicy):
                    _com_interfaces_ = [IID_ISearchCompletedCallback]
                    _reg_clsid_ = '{41B032DA-86B5-4907-A7F7-128E59333101}'
                    _reg_progid_ = "wapt.waptwua"

                    #_dispid_to_func_ = {0:'Invoke'}

                    def __init__(self):
                        self._wrap_(self)

                    def _invokeex_(self, dispid, lcid, wFlags, args, kwargs, serviceProvider):
                        print('WUA Search completed !')
                        return 0

                scb = SearchCallback()
                scb_interface = pythoncom.WrapObject(scb)
                search_job = self.update_searcher().BeginSearch(self.params.filter, scb_interface, None)

                estimated_time = float(self.wapt.read_param('waptwua.last_scan_duration', 300.0))
                if estimated_time < 60.0:
                    estimated_time = 300.0
                tick_every = 1.0
                progress_step = 100.0 * tick_every / estimated_time
                start_time = time.time()
                if self.wapttask:
                    self.wapttask.progress = 0.0
                try:
                    print('Waiting for WUA search to complete')
                    if self.wapttask:
                        self.wapttask.runstatus = 'Waiting for WUA search to complete'
                    while True:
                        if search_job.IsCompleted:
                            break
                        if self.wapttask and self.wapttask.wapt:
                            self.wapttask.wapt.check_cancelled()
                        if self.wapttask:
                            self.wapttask.progress += progress_step
                            self.wapttask.runstatus = 'Searching %0.0ds' % (time.time() - start_time,)
                        time.sleep(tick_every)
                    print('Done searching')
                    self.wapt.write_param('waptwua.last_scan_duration', time.time() - start_time)
                finally:
                    search_job.CleanUp()
                    search_result = self.update_searcher().EndSearch(search_job)

                if self.wapttask:
                    self.wapttask.runstatus = 'Done searching'
                    self.wapttask.progress = 100.0

                self._updates = []
                self._cached_updates = {}
                for update in search_result.Updates:
                    update5 = CastTo(update, 'IUpdate5')
                    self._updates.append(update5)
                self.wapt.write_param('waptwua.status', 'READY')
                print('Updates scan done.')
                self.wapt.write_param('waptwua.error_msg', '')
            except Exception as e:
                if self.wapttask:
                    self.wapttask.runstatus = 'Error scanning updates: %s' % ensure_unicode(e)
                    self.wapttask.progress = 100.0
                self.wapt.write_param('waptwua.status', 'ERROR SCANNING UPDATES')
                logger.error('Error scanning updates: %s ' % ensure_unicode(e))
                self.wapt.write_param('waptwua.error_msg', ensure_unicode(e))
                raise

        return self._updates

    def installed_updates(self):
        """Calculate list of installed Update objects.
        Keep it in cache for current session.

        Returns:
            list of IUpdate
        """
        if self._installed_updates is None:
            print('Looking for installed updates')
            self._installed_updates = []
            searcher = self.update_searcher()
            update_count = searcher.GetTotalHistoryCount()
            if update_count > 1:
                for update in searcher.QueryHistory(0, update_count-1):
                    self._installed_updates.append(CastTo(update, 'IUpdateHistoryEntry2'))

        return self._installed_updates

    def scan_updates_status(self, force=False, update_ids=None):
        """Check all updates and filter out which one should be installed

        """
        print('Scanning with windows updates rules:\n%s' % jsondump(self.params, indent=True))
        installed, pending, discarded = 0, 0, 0

        self.update_wsusscan_cab()

        # XXX There are no clear heuristics to tell whether we can bypass scanning or not.
        # Fail safe until we find better.
        if not force and not self.rescan_needed():
            print('Bypassing scan, no change since last successful scan')
            full_stats = self.stored_updates_localstatus()
            if isinstance(full_stats, dict):
                installed = len([u for u in full_stats['updates'] if u['status'] == 'OK'])
                pending = len([u for u in full_stats['updates'] if u['status'] == 'PENDING'])
                discarded = len([u for u in full_stats['updates'] if u['status'] == 'DISCARDED'])
                self.wapt.update_server_status(force=force)
                return (installed, pending, discarded)

        logger.debug('Scanning installed / not installed Updates')
        for update in self.updates():
            self.store_cached_update_property(update, 'IsDownloaded', update.IsDownloaded)

            if self.wapt.waptwua_enabled:
                if not self.cached_update_property(update, 'IsInstalled'):
                    if not self.params.is_delayed(update) and self.is_allowed(update) and (update_ids is None or [u for u in update_ids if u.startswith(update.Identity.UpdateID)]):
                        # IUpdate : https://msdn.microsoft.com/en-us/library/windows/desktop/aa386099(v=vs.85).aspx
                        # IUpdate2 : https://msdn.microsoft.com/en-us/library/windows/desktop/aa386100(v=vs.85).aspx
                        if update.IsHidden:
                            print('Enable %s : %s' % (update.Identity.UpdateID, update.Title))
                            update.IsHidden = False
                        pending += 1
                    else:
                        if not update.IsHidden:
                            print('Disable %s : %s' % (update.Identity.UpdateID, update.Title))
                            update.IsHidden = True
                        discarded += 1
                else:
                    logger.debug('Already installed %s : %s' % (update.Identity.UpdateID, update.Title))
                    installed += 1

        print('Writing status in local wapt DB')
        self.wapt.write_param('waptwua.last_scan_date', datetime2isodate())
        if not pending:
            print('Status: OK')
            self.wapt.write_param('waptwua.status', 'OK')
        else:
            print('Status: PENDING_UPDATES')
            self.wapt.write_param('waptwua.status', 'PENDING_UPDATES')

        self.wapt.write_param('waptwua.rules_packages', self.rules_packages)

        try:
            _update_history = self.update_history()
            if _update_history is not None:
                self.wapt.write_param('waptwua.update_history', _update_history)
            else:
                _update_history = self.wapt.read_param('waptwua.update_history')
        except Exception as e:
            logger.warning('WARNING: Unable to get windows update history: %s' % ensure_unicode(e))
            # better don't store this than nothing at all.
            # can get com_error when scanning this. ( -2145120257 )
            # self.wapt.delete_param('waptwua.update_history')
            _update_history = self.wapt.read_param('waptwua.update_history')
        self.wapt.write_param('waptwua.updates_localstatus', [self.update_local_status_as_dict(u, _update_history) for u in self.updates()])
        self.wapt.write_param('waptwua.updates', [self.update_as_dict(u) for u in self.updates()])
        self.write_last_successful_scan_signature()

        # send status to wapt server
        logger.debug('Updating workstation status on remote wapt server')

        dict_id_description = {}
        # Dict kb description with wua
        list_update_ids_installed = []
        for u in self.updates():
            if self.update_local_status_as_dict(u)['installed']:
                list_update_ids_installed.append(self.update_local_status_as_dict(u)['update_id'])

        for u in self.updates():
            if self.update_as_dict(u)['update_id'] in list_update_ids_installed:
                dict_id_description['KB%s' % self.update_as_dict(u)['kbids'][0]] = self.update_as_dict(u)['title']

        # update with wmi liste
        for u in installed_windows_updates():
            if not u['HotFixID'] in dict_id_description:
                dict_id_description[u['HotFixID']] = u['Description']

        self.wapt.write_param('waptwua.simple.list', dict_id_description)

        self.wapt.update_server_status(force=force)
        return (installed, pending, discarded)

    def rescan_needed(self):
        """Check if a rescan of Windows updates status is needed.
        * rules have changed
        * wsuscan2 file has been updated
        * updates installation history has changed

        Returns:
            bool: True if rule or updates index or installed updates
        """


        lastscan_params_cksum = self.wapt.read_param('waptwua.params_cksum')

        if lastscan_params_cksum != hashlib.sha1(bytes(jsondump_sorted(self.windows_updates_rules)+jsondump_sorted(self.params), 'utf-8')).hexdigest():
            print('Windows updates rules have been changed')
            return True
        # check if some other tools have installed / uninstalled some QFE
        try:
            update_history = self.update_history()
            update_history_checksum = hashlib.sha1(bytes(jsondump_sorted(update_history), 'utf8')).hexdigest()
            old_cksum = self.wapt.read_param('waptwua.update_history_checksum')
            if update_history_checksum != old_cksum:
                print('Some windows updates have been installed/uninstalled since last scan')
                return True
        except:
            logger.error('Unable to get update_history')
            return True

        # check if MS updates scan file have changed

        if not os.path.exists(self.wsusscn2):
            raise Exception('Unexpected: missing scan file %s' % self.wsusscn2)
        new_hash = sha1_for_file(self.wsusscn2)
        old_hash = self.wapt.read_param('waptwua.wsusscn2_checksum')
        if old_hash != new_hash:
            print('WSUSscan2.cab file has changed since last scan')
        return old_hash != new_hash

    def write_last_successful_scan_signature(self):
        """Write in local db a signature which should change
        if rules or needed updates or installed updates have changed
        """
        cksum = sha1_for_file(self.wsusscn2)

        self.wapt.write_param('waptwua.wsusscn2_checksum', cksum)
        update_history = self.wapt.read_param('waptwua.update_history', ptype='json')
        if not update_history:
            try:
                update_history = self.update_history()
                self.wapt.write_param('waptwua.update_history', update_history)
            except Exception:
                logger.error('Unable to get update_history')
        if update_history:
            cksum = hashlib.sha1(bytes(jsondump_sorted(update_history), 'utf8')).hexdigest()
            self.wapt.write_param('waptwua.update_history_checksum', cksum)

        self.windows_updates_rules.save_to_waptdb(self.wapt)
        self.params.save_to_waptdb(self.wapt)
        self.wapt.write_param('waptwua.params_cksum', hashlib.sha1(bytes(jsondump_sorted(self.windows_updates_rules)+jsondump_sorted(self.params), 'utf8')).hexdigest())

    def wget_update(self, url):
        # try using specialized proxy
        url_parts = urlparse(url)
        target = makepath(self.cache_path, os.path.split(url)[1])
        filename = os.path.split(url)[1]
        repo = self.wuarepo()
        if repo:
            repo_parts = urlparse(repo.repo_url)
        else:
            repo_parts = None

        if repo:
            with repo.get_requests_session() as session:
                # direct download of prefetched cab

                if self.params.direct_download or ('download.windowsupdate.com' in repo.repo_url):
                    print("Downloading %s" % url)
                    wget(url, target,
                         verify_cert=True,
                         resume=True,
                         connect_timeout=self.wuarepo().timeout,
                         sha1=get_sha1_from_filename(url_parts.path),  # MS Cabs have sometime the sha1 in the filename, just before the extension : pciclearstalecache_1d566f78af356cc09c83cf173dc734c66659e0d2.exe
                         proxies=repo.proxies,
                         cache_dir=self.cache_path,
                         limit_bandwidth=self.wapt.limit_bandwidth
                         )

                else:
                    patch_url = '%s://%s%swua/%s' % (repo_parts.scheme, repo_parts.netloc, repo_parts.path, filename)
                    print("Downloading %s" % patch_url)
                    wget(patch_url, target,
                         resume=True,
                         connect_timeout=repo.timeout,
                         sha1=get_sha1_from_filename(url_parts.path),  # MS Cabs have sometime the sha1 in the filename, just before the extension : pciclearstalecache_1d566f78af356cc09c83cf173dc734c66659e0d2.exe
                         cache_dir=self.cache_path,
                         requests_session=session,
                         limit_bandwidth=self.wapt.limit_bandwidth
                         )

        return target

    def _check_disk_space(self):
        if wmi:
            for d in wmi.WMI().Win32_LogicalDisk():
                device = d.Name
                if device == os.environ['SystemDrive'] and int(d.FreeSpace) < 2 ** 30:
                    raise WAPTDiskSpaceException('Not enough space left on device ' + device, d.FreeSpace)

    def get_downloads_for_update(self, update):
        """Returns all the URL to download for the upadte

        Returns:
            list
        """
        result = []
        for dc in update.DownloadContents:
            # https://msdn.microsoft.com/en-us/library/windows/desktop/aa386120(v=vs.85).aspx
            if authorized_url_wsus(dc.DownloadUrl):
                result.append(dc.DownloadUrl)

        for bu in update.BundledUpdates:
            files = win32com_ensure_dispatch_patch('Microsoft.Update.StringColl')
            for dc in bu.DownloadContents:
                # https://msdn.microsoft.com/en-us/library/windows/desktop/aa386120(v=vs.85).aspx
                if authorized_url_wsus(dc.DownloadUrl):
                    result.append(dc.DownloadUrl)
        return result

    def missing_files_in_cache(self, update):
        """Returns the list of URL

        Returns:
            list of str (url)
        """
        result = []
        downloads = self.get_downloads_for_update(update)
        for url in downloads:
            target = makepath(self.cache_path, os.path.split(url)[1])
            if not os.path.isfile(target):
                result.append(url)
        return result

    def download_single(self, update):
        """Download the files associated to an IUpdate instance
        Direct content
        Bundled content

        Args:
            update (IUpdate): update

        Returns:
            dict(downloaded=result,missing=missing_urls)

        """

        def purge_waptwua_cache(files):
            """Removes files from wapt\cache\waptwua"""
            for fn in list(files):
                try:
                    os.remove(fn)
                except Exception as e:
                    logger.error('Unable to delete cached file %s' % fn)

        result = []
        missing_urls = []

        self._check_disk_space()

        update = CastTo(update, 'IUpdate5')

        for dc in update.DownloadContents:
            # https://msdn.microsoft.com/en-us/library/windows/desktop/aa386120(v=vs.85).aspx
            files = win32com_ensure_dispatch_patch('Microsoft.Update.StringColl')
            try:
                print("Downloading %s" % dc.DownloadUrl)
                target = self.wget_update(dc.DownloadUrl)
            except Exception as e:
                missing_urls.append(dc.DownloadUrl)
                logger.error("ERROR: skipping download %s, reason: %s" % (dc.DownloadUrl, ensure_unicode(e)))
                continue

            result.append(dc.DownloadUrl)
            if isfile(target):
                files.Add(target)

            for fn in files:
                print("%s will be put to local WUA cache for update" % (fn,))
                # if isfile(fn):
                #    remove_file(fn)

            if len(list(files)) > 0:
                try:
                    print('Copying %s downloaded files to MS Wua cache' % files.Count)
                    update.CopyToCache(files)
                    purge_waptwua_cache(files)
                except Exception as e:
                    logger.error("ERROR: unable to push some update files in Windows cache reason: %s" % (ensure_unicode(e),))

        for bu in update.BundledUpdates:
            bu = CastTo(bu, 'IUpdate5')
            files = win32com_ensure_dispatch_patch('Microsoft.Update.StringColl')
            for dc in bu.DownloadContents:
                # https://msdn.microsoft.com/en-us/library/windows/desktop/aa386120(v=vs.85).aspx
                try:
                    target = self.wget_update(dc.DownloadUrl)
                    result.append(dc.DownloadUrl)
                except Exception as e:
                    logger.error("ERROR: skipping download %s, reason: %s" % (dc.DownloadUrl, ensure_unicode(e)))
                    missing_urls.append(dc.DownloadUrl)
                    continue
                if isfile(target):
                    files.Add(target)

            for fn in files:
                print("%s put to local WUA cache for update" % (fn,))

            if len(list(files)) > 0:
                try:
                    print('Copying %s downloaded files to MS Wua cache' % files.Count)
                    bu.CopyToCache(files)
                    purge_waptwua_cache(files)
                except Exception as e:
                    logger.error("ERROR: unable to push update files in Windows cache, reason: %s" % (ensure_unicode(e),))

        if missing_urls:
            self.store_cached_update_property(update, 'IsDownloaded', False)
        else:
            self.store_cached_update_property(update, 'IsDownloaded', True)

        return dict(downloaded=result, missing=missing_urls)

    def missing_downloads(self):
        to_install = [u['update_id'] for u in self.stored_updates_localstatus() if not u['installed'] and not u['downloaded'] and u['allowed']]
        if to_install:
            updates_urls = {u['update_id']: u['download_urls'] for u in self.stored_updates()}
            result = []
            for u in to_install:
                urls = updates_urls[u]
                for url in urls:
                    if not url in result:
                        result.append(url)
            return result
        else:
            return None

    def download_updates(self, force=False, uuids=None):
        """Download all pending updates and put them in Windows Update cache

        Args:
            force (bool) : download update even if update is tagged as donwloaded.
            uuids (list of str) : list of update_id for which to download files.
                                  if None, download all missing files for allowed not installed updates.

        Returns:
            dict (downloaded=[], missing=[])

        """
        self.update_wsusscan_cab()
        result = dict(downloaded=[], missing=[])
        try:
            missing = self.missing_downloads()
            if force or missing is None or self.rescan_needed() or uuids:
                for update in self.updates():
                    if (force or not update.IsInstalled) and self.is_allowed(update) and (force or not update.IsDownloaded) and (uuids is None or [u for u in uuids if u.startswith(update.Identity.UpdateID)]):
                        # IUpdate : https://msdn.microsoft.com/en-us/library/windows/desktop/aa386099(v=vs.85).aspx
                        # IUpdate2 : https://msdn.microsoft.com/en-us/library/windows/desktop/aa386100(v=vs.85).aspx
                        res = self.download_single(update)
                        result['downloaded'].extend(res['downloaded'])
                        result['missing'].extend([url for url in res['missing'] if url not in result['missing']])
                self.scan_updates_status()
            elif missing is not None:
                # download only in wapt cache
                missing_urls = []
                for url in missing:
                    try:
                        print("Downloading %s" % url)
                        target = self.wget_update(url)
                    except Exception as e:
                        missing_urls.append(url)
                        logger.error("ERROR: skipping download %s, reason: %s" % (url, ensure_unicode(e)))
                        continue
            self.wapt.write_param('waptwua.missing_downloads', result['missing'])
            self.wapt.write_param('waptwua.error_msg', '')
        except WAPTDiskSpaceException:
            logger.error('Disk Space Error')
            self.wapt.write_param('waptwua.last_error_msg', 'Disk Space Error')
            self.wapt.write_param('waptwua.status', 'DISK_SPACE_ERROR')
        except Exception as e:
            logger.error('Unexpected error: %s' % ensure_unicode(e))
            self.wapt.write_param('waptwua.last_error_msg', ensure_unicode(e))
            self.wapt.write_param('waptwua.status', 'ERROR')
            self.wapt.write_param('waptwua.error_msg', ensure_unicode(e))
        self.wapt.update_server_status()
        return result

    def install_updates(self, force=False, uuids=None):
        """Install all pending allowded updates or updates specified in uuids arg.
        If update is not yet downloaded, a downloaded is triggered.
        Results are stored in local db and sent to server database.

        Args:
            force (bool) : install update even if update is tagged as installed.
            uuids (list of str) : list of update_id to install.
                                  if None, all allowed not installed updates are installed.

        Returns:
            list of str: list of UpdateId of triggered installed updates

        """
        result = []
        missing_downloads = []
        self.update_wsusscan_cab()

        if uuids is not None:
            print('Start of Windows updates install for: %s' % uuids)
        else:
            print('Start of install for all pending Windows updates')
        print('Scanning with params:\n%s' % jsondump(self.params, indent=True))
        print('Scanning with windows updates rules:\n%s' % jsondump(self.windows_updates_rules, indent=True))

        discarded = 0
        try:
            self.wapt.write_param('waptwua.status', 'INSTALL')
            updates_to_install = win32com_ensure_dispatch_patch("Microsoft.Update.UpdateColl")
            updates_to_install_list = []
            # apply the updates
            for update in self.updates():
                if self.is_allowed(update) and (uuids is None or [u for u in uuids if u.startswith(update.Identity.UpdateID)]) and not update.IsInstalled:
                    if not update.IsDownloaded:
                        res = self.download_single(update)
                        if res['missing']:
                            missing_downloads.extend(res['missing'])
                            logger.warning('Missing update packages: %s' % res['missing'])
                            discarded += 1
                            continue
                    print('Adding %s to the list of updates to install...' % update.Identity.UpdateID)
                    update.AcceptEula()
                    update.IsHidden = False
                    updates_to_install.Add(update)
                    updates_to_install_list.append(update)
                    result.append("%s_%s" % (update.Identity.UpdateID, update.Identity.RevisionNumber))
                elif not update.IsInstalled and not self.is_allowed(update) and self.wapt.waptwua_enabled:
                    print('Denied %s : %s' % (update.Identity.UpdateID, update.Title))
                    if not update.IsHidden:
                        update.IsHidden = True
                    discarded += 1
                else:
                    print('Skipped %s : %s' % (update.Identity.UpdateID, update.Title))
                    discarded += 1

            if result:
                start_install_time = time.time()
                installer = self.update_session().CreateUpdateInstaller()
                installer.Updates = updates_to_install
                if installer.RebootRequiredBeforeInstallation:
                    raise EWUARebootNeeded('Reboot is needed before installing Windows updates')
                installer.AllowSourcePrompts = False
                installer.IsForced = force
                print('Launching the install of %s updates' % len(result))

                IID_IInstallationProgressChangedCallback = '{E01402D5-F8DA-43BA-A012-38894BD048F1}'
                IID_IInstallationCompletedCallback = '{45F4F6F3-D602-4F98-9A8A-3EFA152AD2D3}'

                class InstallationProgressChangedCallback(EventHandlerPolicy):
                    _public_methods_ = []
                    _com_interfaces_ = [IID_IInstallationProgressChangedCallback]

                    _reg_clsid_ = '{41B032DA-86B5-4907-A7F7-128E59333011}'
                    _reg_progid_ = "wapt.waptwua"

                    def __init__(self, wapt):
                        self._wrap_(self)
                        self.wapt = wapt

                    def _invokeex_(self, dispid, lcid, wFlags, args, kwargs, serviceProvider):
                        try:
                            installJob = win32com_ensure_dispatch_patch(args[0])
                            callbackArgs = win32com_ensure_dispatch_patch(args[1])
                            local_progress = callbackArgs.Progress
                            current_progress = local_progress.CurrentUpdateIndex+1
                            max_progress = len(result)

                            msg = 'Installing update %s / %s %s%% completed' % (current_progress, max_progress, local_progress.PercentComplete)

                            if self.wapttask and max_progress:
                                self.wapttask.progress = current_progress * 100.0 / max_progress
                                self.wapttask.runstatus = msg

                            if self.wapt.progress_hook is None:
                                print(msg)
                            else:
                                cancel_request = self.wapt.progress_hook(True, msg, current_progress, max_progress)
                                if cancel_request:
                                    print('Cancel Windows updates install requested')
                                    installJob.RequestAbort()

                        except:
                            pass

                        return 0

                class InstallationCompletedCallback(DesignatedWrapPolicy):
                    _public_methods_ = []
                    _com_interfaces_ = [IID_IInstallationCompletedCallback]

                    _reg_clsid_ = '{41B032DA-86B5-4907-A7F7-128E59333013}'
                    _reg_progid_ = "wapt.waptwua"

                    def __init__(self):
                        self._wrap_(self)

                    # dispid, lcid, wFlags, args
                    def _invokeex_(self, dispid, lcid, wFlags, args, kwargs, serviceProvider):
                        print('Install completed !')
                        return 0

                scb = InstallationProgressChangedCallback(self.wapt)
                scb_com = pythoncom.WrapObject(scb)

                dcb = InstallationCompletedCallback()
                dcb_com = pythoncom.WrapObject(dcb)

                # TODO: calculate estimated time
                estimated_time = float(self.wapt.read_param('waptwua.last_install_duration', 600.0))
                if estimated_time < 60.0:
                    estimated_time = 600.0
                tick_every = 1.0
                progress_step = 100.0 * tick_every / estimated_time
                start_time = time.time()

                install_job = installer.BeginInstall(scb_com, dcb_com, None)
                try:
                    n = 0
                    print('Waiting for install to complete', end='')
                    while True:
                        if install_job.IsCompleted:
                            break
                        if self.wapttask and self.wapttask.wapt:
                            self.wapttask.wapt.check_cancelled()
                        if self.wapttask:
                            self.wapttask.progress += progress_step
                            self.wapttask.runstatus = 'Installing %0.0ds' % (time.time() - start_time,)
                        time.sleep(tick_every)

                    print('Done installing')
                finally:
                    install_job.CleanUp()
                    installation_result = installer.EndInstall(install_job)

                print("Install result: %s" % InstallResult.get(installation_result.ResultCode, installation_result.ResultCode))
                print("Reboot required: %s" % installation_result.RebootRequired)
                self.write_install_status_for_windows(installation_result.ResultCode)
                self.wapt.write_param('waptwua.last_install_duration', time.time() - start_install_time)
                self.wapt.write_param('waptwua.last_install_reboot_required', installation_result.RebootRequired)
                self.wapt.write_param('waptwua.last_install_result', InstallResult[installation_result.ResultCode])
                if installation_result.ResultCode in [3, 4, 5]:
                    self.wapt.write_param('waptwua.status', 'ERROR')
                else:
                    # assume all is installed for the next report...
                    for update in updates_to_install_list:
                        self.store_cached_update_property(update, 'IsInstalled', True)

            else:
                self.wapt.write_param('waptwua.last_install_reboot_required', False)
                self.wapt.write_param('waptwua.last_install_result', None)

            self.wapt.write_param('waptwua.last_install_batch', result)
            self.wapt.write_param('waptwua.last_install_date', datetime2isodate())
            self.wapt.write_param('waptwua.error_msg', '')
        except WAPTDiskSpaceException:
            self.wapt.write_param('waptwua.status', 'DISK_SPACE_ERROR')
            self.wapt.write_param('waptwua.error_msg', 'Not enough free disk space')
            logger.error('Disk Space Error')
        except Exception as e:
            self.wapt.write_param('waptwua.status', 'ERROR')
            self.wapt.write_param('waptwua.error_msg', ensure_unicode(e))
            logger.error("ERROR installing update: %s" % ensure_unicode(e))
        finally:
            self.scan_updates_status()
        self.wapt.update_server_status()
        return result

    def uninstall_updates(self, force=False, uuids=[]):
        """Uninstall a liste of updates specified by its uuids arg.

        Args:
            force (bool) : install update even if update is tagged as installed.
            uuids (list of str) : list of update_id to uninstall.

        Returns:
            list of str: list of UpdateId of triggered installed updates

        """
        result = []
        missing_downloads = []
        self.update_wsusscan_cab()

        print('Start of Windows updates uninstall for: %s' % uuids)

        print('Scanning with params:\n%s' % jsondump(self.params, indent=True))
        print('Scanning with windows updates rules:\n%s' % jsondump(self.windows_updates_rules, indent=True))

        discarded = 0
        try:
            self.wapt.write_param('waptwua.status', 'INSTALL')
            updates_to_uninstall = win32com_ensure_dispatch_patch("Microsoft.Update.UpdateColl")
            updates_to_uninstall_list = []

            # apply the updates
            for update in self.updates():
                if [u for u in uuids if u.startswith(update.Identity.UpdateID)] and update.IsInstalled:
                    print('Adding %s to the list of updates to uninstall...' % update.Identity.UpdateID)
                    update.IsHidden = True
                    # updates_to_uninstall.Add(update)
                    updates_to_uninstall_list.append(update)
                    result.append("%s_%s" % (update.Identity.UpdateID, update.Identity.RevisionNumber))

            if result:
                start_install_time = time.time()
                errors = []

                for update in updates_to_uninstall_list:
                    with disable_file_system_redirection():
                        try:
                            kbids = ["%s" % kb for kb in update.KBArticleIDs]
                            for kb in kbids:
                                run('wusa.exe /kb:%s /uninstall /quiet /norestart' % kb)
                            self.store_cached_update_property(update, 'IsInstalled', False)
                        except Exception as e:
                            error = '%s / %s : %s' % (update.Identity.UpdateID, kbids, ensure_unicode(e))
                            errors.append(error)
                            logger.warning('Unable to uninstall %s' % error)

                    if errors:
                        self.wapt.write_param('waptwua.status', 'ERROR')
                        self.wapt.write_param('waptwua.error_msg', errors)

        except WAPTDiskSpaceException:
            self.wapt.write_param('waptwua.status', 'DISK_SPACE_ERROR')
            self.wapt.write_param('waptwua.error_msg', 'Not enough free disk space')
            logger.error('Disk Space Error')
        except Exception as e:
            self.wapt.write_param('waptwua.status', 'ERROR')
            self.wapt.write_param('waptwua.error_msg', ensure_unicode(e))
            logger.error("ERROR uninstalling update: %s" % ensure_unicode(e))
        finally:
            self.scan_updates_status()
        self.wapt.update_server_status()
        return result

    def stored_waptwua_status(self):
        """Returns the status of Wapt WUA agent as reported in local WaptDB by the last scan operation

        Returns:
            dict
        """
        return {
            'enabled': self.wapt.waptwua_enabled,
            'status': self.wapt.read_param('waptwua.status', None),
            'wsusscn2cab_date': self.wapt.read_param('waptwua.wsusscn2cab_date', None),
            'last_scan_date': self.wapt.read_param('waptwua.last_scan_date', None),
            'last_scan_duration': self.wapt.read_param('waptwua.last_scan_duration', None),
            'last_install_batch': self.wapt.read_param('waptwua.last_install_batch', [], ptype='json'),
            'last_install_date': self.wapt.read_param('waptwua.last_install_date', None),
            'last_install_result': self.wapt.read_param('waptwua.last_install_result', None),
            'last_install_reboot_required': self.wapt.read_param('waptwua.last_install_rebootrequired', None),
            'missing_downloads': self.wapt.read_param('waptwua.missing_downloads', None, ptype='json'),
            'rules_packages': self.wapt.read_param('waptwua.rules_packages', None, ptype='json'),
            'last_error': self.wapt.read_param('waptwua.error_msg')
        }

    def stored_waptwua_rules(self):
        """Returns the status of Wapt WUA agent as reported in local WaptDB by the last scan operation

        Returns:
            dict
        """
        return {
            'windows_updates_rules': WaptWUARules().load_from_waptdb(self.wapt).as_dict(),
            'params': WaptWUAParams().load_from_waptdb(self.wapt).as_dict(),
        }

    def stored_updates_localstatus(self):
        """Returns the list of Windows updates with their status for this computer
        as reported in local DB by the last scan/download/install operation.

        Returns:
            list
        """
        return self.wapt.read_param('waptwua.updates_localstatus', [], ptype='json')

    def stored_updates(self):
        """Returns the list of potentially applicable Windows updates for this computer
        with details extracted from Wsusscan2 cab file by the IUpdateSearcher
        Windows process.

        Returns:
            list
        """
        return self.wapt.read_param('waptwua.updates', [], ptype='json')

    def write_scan_status_for_windows(self, last_error_code=0,):
        """Update windows configuration panel registry with update result

        [HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows\CurrentVersion\WindowsUpdate\Auto Update\Results\Detect]
        "LastSuccessTime"="2016-12-23 15:59:32"
        "LastError"=dword:00000000
        """
        registry_set(HKEY_LOCAL_MACHINE, r'SOFTWARE\Microsoft\Windows\CurrentVersion\WindowsUpdate\Auto Update\Results\Detect', "LastError", last_error_code, REG_DWORD)
        registry_set(HKEY_LOCAL_MACHINE, r'SOFTWARE\Microsoft\Windows\CurrentVersion\WindowsUpdate\Auto Update\Results\Detect', "LastSuccessTime", datetime_for_reg())

    def write_download_status_for_windows(self, last_error_code=0,):
        """Update windows configuration panel registry with update result
        [HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows\CurrentVersion\WindowsUpdate\Auto Update\Results\Download]
        "LastSuccessTime"="2016-11-10 16:24:29"
        "LastError"=dword:00000000
        """
        registry_set(HKEY_LOCAL_MACHINE, r'SOFTWARE\Microsoft\Windows\CurrentVersion\WindowsUpdate\Auto Update\Results\Download', "LastError", last_error_code, REG_DWORD)
        registry_set(HKEY_LOCAL_MACHINE, r'SOFTWARE\Microsoft\Windows\CurrentVersion\WindowsUpdate\Auto Update\Results\Download', "LastSuccessTime", datetime_for_reg())

    def write_install_status_for_windows(self, last_error_code=0,):
        """Update windows configuration panel registry with update result
        [HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows\CurrentVersion\WindowsUpdate\Auto Update\Results\Install]
        "LastSuccessTime"="2016-11-11 02:00:43"
        "LastError"=dword:00000000
        """
        registry_set(HKEY_LOCAL_MACHINE, r'SOFTWARE\Microsoft\Windows\CurrentVersion\WindowsUpdate\Auto Update\Results\Install', "LastError", last_error_code, REG_DWORD)
        registry_set(HKEY_LOCAL_MACHINE, r'SOFTWARE\Microsoft\Windows\CurrentVersion\WindowsUpdate\Auto Update\Results\Install', "LastSuccessTime", datetime_for_reg())

    def disable_ms_windows_update_service(self):
        """Set windows update service in disabled state and disable external Windows Update source"""
        """[HKEY_LOCAL_MACHINE\SOFTWARE\Policies\Microsoft\Windows\WindowsUpdate]
        "DisableOSUpgrade"=dword:00000001
        "DisableWindowsUpdateAccess"=dword:00000001

        [HKEY_LOCAL_MACHINE\SYSTEM\Internet Communication Management\Internet Communication]
        "DisableWindowsUpdateAccess"=dword:00000001

        [HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion\Policies\WindowsUpdate]
        "DisableWindowsUpdateAccess"=dword:00000001

        [HKEY_LOCAL_MACHINE\Software\Policies\Microsoft\Windows\WindowsUpdate\AU]
        "NoAutoUpdate"=dword:00000001

        [HKEY_LOCAL_MACHINE\Software\Policies\Microsoft\Windows\WindowsUpdate\AU]
        "UseWUServer"=dword:00000001
        """

        registry_set(HKEY_LOCAL_MACHINE, r'SOFTWARE\Policies\Microsoft\Windows\WindowsUpdate', 'DisableOSUpgrade', 1, REG_DWORD)
        registry_set(HKEY_LOCAL_MACHINE, r'SOFTWARE\Policies\Microsoft\Windows\Gwx', 'DisableGwx', 1, REG_DWORD)
        registry_set(HKEY_LOCAL_MACHINE, r'SOFTWARE\Policies\Microsoft\Windows\WindowsUpdate', 'DisableWindowsUpdateAccess', 1, REG_DWORD)

        registry_set(HKEY_LOCAL_MACHINE, r'SYSTEM\Internet Communication Management\Internet Communication', 'DisableWindowsUpdateAccess', 1, REG_DWORD)
        registry_set(HKEY_LOCAL_MACHINE, r'Software\Microsoft\Windows\CurrentVersion\Policies\WindowsUpdate', 'DisableWindowsUpdateAccess', 1, REG_DWORD)
        registry_set(HKEY_LOCAL_MACHINE, r'Software\Microsoft\Windows\CurrentVersion\Policies\WindowsUpdate', 'UseWUServer', 1, REG_DWORD)

        if wmi:
            c = wmi.WMI()
            for service in c.Win32_Service(Name='wuauserv'):
                # https://msdn.microsoft.com/en-us/library/aa384896(v=vs.85).aspx
                service.ChangeStartMode(StartMode="Disabled")
                # don't stop if in contextual EnsureWUAUServRunning
                if service.Started and (not hasattr(self, 'old_au_options') or self.old_au_options is None):
                    service.StopService()

    def enable_ms_windows_update_service(self):
        registry_set(HKEY_LOCAL_MACHINE, r'SOFTWARE\Policies\Microsoft\Windows\WindowsUpdate', 'DisableOSUpgrade', 0, REG_DWORD)
        registry_set(HKEY_LOCAL_MACHINE, r'SOFTWARE\Policies\Microsoft\Windows\WindowsUpdate', 'DisableWindowsUpdateAccess', 0, REG_DWORD)

        registry_set(HKEY_LOCAL_MACHINE, r'SYSTEM\Internet Communication Management\Internet Communication', 'DisableWindowsUpdateAccess', 0, REG_DWORD)
        registry_set(HKEY_LOCAL_MACHINE, r'Software\Microsoft\Windows\CurrentVersion\Policies\WindowsUpdate', 'DisableWindowsUpdateAccess', 0, REG_DWORD)
        registry_set(HKEY_LOCAL_MACHINE, r'Software\Microsoft\Windows\CurrentVersion\Policies\WindowsUpdate', 'UseWUServer', 0, REG_DWORD)

        # https://docs.microsoft.com/fr-fr/security-updates/windowsupdateservices/18127152
        # Range = 2|3|4|5
        #   2 = Notify before download.
        #   3 = Automatically download and notify of installation.
        #   4 = Automatic download and scheduled installation. (Only valid if values exist for ScheduledInstallDay and ScheduledInstallTime.)
        #   5 = Automatic Updates is required, but end users can configure it.
        registry_set(HKEY_LOCAL_MACHINE, r'SOFTWARE\Microsoft\Windows\CurrentVersion\WindowsUpdate\Auto Update', 'AUOptions', 2, REG_DWORD)

        if wmi:
            c = wmi.WMI()
            for service in c.Win32_Service(Name='wuauserv'):
                # https://msdn.microsoft.com/en-us/library/aa384896(v=vs.85).aspx
                service.ChangeStartMode(StartMode="Automatic")
                # don't restart if in contextual EnsureWUAUServRunning
                if not service.Started and (not hasattr(self, 'old_au_options') or self.old_au_options is None):
                    service.StartService()

    # nom_params_serveur = ['WUServer','WUStatusServer']
    # tableau des variables de config et de la valeur attribuÃƒÆ’Ã‚Â©e
    # nom_params_config = [('NoAutoRebootWithLoggedOnUsers',1),('UseWUServer',1),('RescheduleWaitTime',5),('NoAutoUpdate',0),('AUOptions',4),('ScheduledInstallDay',0),('ScheduledInstallTime',wsusheure)]

    def get_wuauserv_status(self):
        return dict(
            start_mode=registry_readstring(HKEY_LOCAL_MACHINE, r'SYSTEM\CurrentControlSet\services\wuauserv', 'Start', None),
            running=service_is_running('wuauserv'),
            auto_update=registry_readstring(HKEY_LOCAL_MACHINE, r'SOFTWARE\Microsoft\Windows\CurrentVersion\WindowsUpdate\Auto Update', 'AUOptions'),
            use_wuserver=registry_readstring(HKEY_LOCAL_MACHINE, r'Software\Microsoft\Windows\CurrentVersion\Policies\WindowsUpdate', 'UseWUServer', None),
            disable_os_upgrade=registry_readstring(HKEY_LOCAL_MACHINE, r'SOFTWARE\Policies\Microsoft\Windows\WindowsUpdate', 'DisableOSUpgrade', None),
            disable_access=registry_readstring(HKEY_LOCAL_MACHINE, r'Software\Microsoft\Windows\CurrentVersion\Policies\WindowsUpdate', 'DisableWindowsUpdateAccess', None),
            policy_disable_access=registry_readstring(HKEY_LOCAL_MACHINE, r'SOFTWARE\Policies\Microsoft\Windows\WindowsUpdate', 'DisableWindowsUpdateAccess', None),
            gwx_disable=registry_readstring(HKEY_LOCAL_MACHINE, r'SOFTWARE\Policies\Microsoft\Windows\Gwx', 'DisableGwx', None),
            reboot_needed=waiting_for_reboot(),
            agent_version=get_file_properties(makepath(system32(), 'wuapi.dll'))['ProductVersion'],
        )

    def summary_status(self):
        stat = self.stored_waptwua_status()
        updates_localstatus = self.stored_updates_localstatus()

        stat['installed'] = len([u for u in updates_localstatus if u['status'] == 'OK'])
        stat['pending'] = len([u for u in updates_localstatus if u['status'] == 'PENDING'])
        stat['discarded'] = len([u for u in updates_localstatus if u['status'] == 'DISCARDED'])
        if stat['pending'] > 0:
            stat['status'] = stat['status']+', %s pending' % stat['pending']
        if stat['discarded'] > 0:
            stat['status'] = stat['status']+', %s discarded' % stat['discarded']

        return """\
Status:            %(status)s

Installed updates: %(installed)s
Pending updates:   %(pending)s
Discarded updates: %(discarded)s
Reboot required:   %(last_install_reboot_required)s

Last install date:   %(last_install_date)s
Last install result: %(last_install_result)s

WSUSScan cab date: %(wsusscn2cab_date)s
        """ % stat


def summary_waptwua_status(wapt):
    wua = WaptWUA(wapt)
    return wua.summary_status()


if __name__ == '__main__':
    # we import only if called as a command line utility.
    # else circular reference.
    from common import Wapt, WaptRepo

    def_allowed_updates = None
    def_forbidden_updates = None
    def_allowed_severities = None
    def_allowed_classifications = None

    parser = OptionParser(usage=__doc__)
    parser.add_option("-S", "--severities", dest="allowed_severities", default=def_allowed_severities, help="Allow updates by severity. csv list of Critical,Important,Moderate,Low. If empty : allow all. (default: %default)")
    parser.add_option("-C", "--classifications", dest="allowed_classifications", default=def_allowed_classifications, help="Allow updates by claffication. csv list of "+','.join(list(UpdateClassifications.values()))+". If empty : allow all. (default: %default)")
    parser.add_option("-a", "--allowed", dest="allowed_updates", default=def_allowed_updates, help="Allow updates by update-id or KB. csv list of id to allow (default: %default)")
    parser.add_option("-b", "--forbidden", dest="forbidden_updates", default=def_forbidden_updates, help="Forbid updates by update-id or KB. csv list (default: %default)")
    parser.add_option("-c", "--config", dest="config", default=None, help="Config file full path (default: %default)")
    parser.add_option("-l", "--loglevel", dest="loglevel", default=None, type='choice',  choices=['debug', 'warning', 'info', 'error', 'critical'], metavar='LOGLEVEL', help="Loglevel (default: warning)")
    #parser.add_option("-d","--dry-run", dest="dry_run",    default=False, action='store_true', help="Dry run (default: %default)")
    parser.add_option("-f", "--force", dest="force", default=False, action='store_true', help="Force scan even if last scan signature has not changed (default: %default)")

    (options, args) = parser.parse_args()

    logging.basicConfig(format='%(asctime)s [%(name)-15s] %(levelname)s %(message)s')

    def setloglevel(logger, loglevel):
        """set loglevel as string"""
        if loglevel in ('debug', 'warning', 'info', 'error', 'critical'):
            numeric_level = getattr(logging, loglevel.upper(), None)
            if not isinstance(numeric_level, int):
                raise ValueError('Invalid log level: {}'.format(loglevel))
            logger.setLevel(numeric_level)

    # force loglevel
    if options.loglevel is not None:
        setloglevel(logger, options.loglevel)

    wapt = Wapt(config_filename=options.config)

    allowed_updates = ensure_list(options.allowed_updates, allow_none=True)
    forbidden_updates = ensure_list(options.forbidden_updates, allow_none=True)
    allowed_severities = ensure_list(options.allowed_severities, allow_none=True)
    allowed_classifications = ensure_list(options.allowed_classifications, allow_none=True)

    windows_updates_rules = WaptWUARules()
    windows_updates_rules.load_from_ini(config=wapt.config)
    windows_updates_rules.load_from_options(options)

    if len(args) < 1:
        print(parser.usage)
        sys.exit(1)

    action = args[0]

    with WaptWUA(wapt, windows_updates_rules=windows_updates_rules) as wua:
        wua.ensure_minimum_wua_version()
        if action == 'scan':
            installed, pending, discarded = wua.scan_updates_status(force=options.force)
            print(summary_waptwua_status(wapt))
        elif action == 'download':
            print(wua.download_updates(force=options.force))
            print(summary_waptwua_status(wapt))
        elif action == 'install':
            print(wua.install_updates(force=options.force))
            print(summary_waptwua_status(wapt))
        elif action == 'status':
            print(summary_waptwua_status(wapt))
        else:
            print(parser.usage)
